import re

from .constants import (
    MODULO10_QUANTIDADE_MOEDA,
    MODULO10_VALOR_EFETIVO,
    MODULO11_QUANTIDADE_MOEDA,
    MODULO11_VALOR_EFETIVO,
)


def calc_dac_modulo10(sequencia: str) -> int:
    """
    O DAC (Dígito de Auto-Conferência) módulo 10, de um número é calculado
    multiplicando cada algarismo, pela seqüência de multiplicadores 2, 1,
    2, 1, ... posicionados da direita para a esquerda.

    A soma dos algarismos do produto é dividida por 10 e o DAC será a diferença
    entre o divisor (10) e o resto da divisão:

        DAC = 10 - (resto da divisão)

    Observação: quando o resto da divisão for 0 (zero), o DAC calculado é o
    0 (zero).

    EXEMPLO

    Calcular o DAC módulo 10 da seguinte seqüência de números: 01230067896.
    A fórmula do cálculo é:

        1. Multiplicação pela sequência 2, 1, 2, 1, ... da direita para a
        esquerda.
              0  1  2  3  0  0  6  7  8  9  6
            X 2  1  2  1  2  1  2  1  2  1  2
              0  1  4  3  0  0 12  7 16  9 12

        2. Soma dos dígitos do produto
            0 + 1 + 4 + 3 + 0 + 0 + 1 + 2 + 7 + 1 + 6 + 9 + 1 + 2 = 37
            Observação: Cada dígito deverá ser somado individualmente.

        3. Divisão do resultado da soma acima por 10
            37 : 10 = 3 , resto = 7
            DAC = 10 - (resto da divisão), portando 10 - 7 = 3

    O DAC da seqüência numérica é igual a “3”.
    """
    multiplicador_atual = 2
    soma = 0

    for element in sequencia[::-1]:
        resultado = int(element) * multiplicador_atual
        multiplicador_atual = 1 if multiplicador_atual == 2 else 2
        if resultado < 10:
            soma += resultado
        else:
            soma += int(resultado / 10)
            soma += resultado % 10

    resto_divisao = soma % 10
    if resto_divisao == 0:
        return resto_divisao
    return 10 - resto_divisao


def calc_dac_modulo11(sequencia) -> int:
    """
    O DAC (Dígito de Auto-Conferência) módulo 11, de um número é calculado
    multiplicando cada algarismo, pela seqüência de multiplicadores 2,3,4,5,6,
    7,8,9,2,3,4....
    posicionados da direita para a esquerda.

    A soma dos produtos dessa multiplicação é dividida por 11, obtém-se o
    resto da divisão, este resto deve ser subtraído de 11, o produto da
    subtração é o DAC.

    Observação: Quando o resto da divisão for igual a 0 ou 1, atribuí-se ao DV
    o digito “0”, e quando for 10, atribuí-se ao DV o digito “1”.

    EXEMPLO

    Calcular o DAC módulo 11 da seguinte seqüência de números: 01230067896.
    A fórmula do cálculo é:

        1. Multiplicação pela seqüência 2,3,4,5,6,7,8,9, 2,3,4 da direita
        para a esquerda.
              0  1  2  3  0  0  6  7  8  9  6
            X 4  3  2  9  8  7  6  5  4  3  2
              0  3  4 27  0  0 36 35 32 27 12

        2. Soma dos produtos da multiplicação:
            0 + 3 + 4 + 27 + 0 + 0 + 36 + 35 + 32 + 27 + 12 = 176

        3. Divisão do resultado da soma acima por 11
            176 : 11 = 16 , resto = 0

    O DAC da seqüência numérica é igual a “0”.
    """
    multiplicador_atual = 2
    soma = 0

    for element in sequencia[::-1]:
        resultado = int(element) * multiplicador_atual
        multiplicador_atual = (
            2 if multiplicador_atual == 9 else multiplicador_atual + 1
        )
        soma += resultado

    resto_divisao = soma % 11
    if resto_divisao in (0, 1):
        return 0
    if resto_divisao == 10:
        return 1
    return 11 - resto_divisao


def dac_func(codigo_moeda):
    modulo = None
    if (
        int(codigo_moeda) == MODULO10_VALOR_EFETIVO
        or int(codigo_moeda) == MODULO10_QUANTIDADE_MOEDA
    ):
        modulo = calc_dac_modulo10
    elif (
        int(codigo_moeda) == MODULO11_VALOR_EFETIVO
        or int(codigo_moeda) == MODULO11_QUANTIDADE_MOEDA
    ):
        modulo = calc_dac_modulo11
    return modulo


def limpar_codigo_de_barras(codigo_de_barras: str):
    return ''.join([i for i in re.findall(r'\d+', codigo_de_barras)])



PRODUTO_ARRECADACAO = 8

SEGMENTO_PREFEITURA = 1
SEGMENTO_SANEAMENTO = 2
SEGMENTO_ENERGIA_ELETRICA_GAS = 3
SEGMENTO_TELECOMUNICACOES = 4
SEGMENTO_ORGAOS_GOVERNAMENTAIS = 5
SEGMENTO_DEMAIS = 6
SEGMENTO_MULTAS_TRANSITO = 7
SEGMENTO_EXCLUSIVO_BANCO = 9

MODULO10_VALOR_EFETIVO = 6
MODULO10_QUANTIDADE_MOEDA = 7
MODULO11_VALOR_EFETIVO = 8
MODULO11_QUANTIDADE_MOEDA = 9

ARRECADACAO_GUICHE_COM_FATURA = '1'
"""1 – Guichê de Caixa com fatura/guia de arrecadação"""

ARRECADACAO_ELETRONICA_COM_FATURA = '2'
"""
2 – Arrecadação Eletrônica com fatura/guia de arrecadação (terminais de
auto - atendimento, ATM, home/office banking)
"""

ARRECADACAO_INTERNET_COM_FATURA = '3'
"""3 – Internet com fatura/guia de arrecadação"""

ARRECADACAO_OUTROS_COM_FATURA = '4'
"""4 – Outros meios com fatura/guia de arrecadação"""

ARRECADACAO_CORRESPONDENTES_COM_FATURA = '5'
"""5 – Correspondentes bancários com fatura/guia de arrecadação"""

ARRECADACAO_TELEFONE_COM_FATURA = '6'
"""6 – Telefone com fatura/guia de arrecadação"""

ARRECADACAO_LOTERICAS_COM_FATURA = '7'
"""7 – Casas lotéricas com fatura/guia de arrecadação"""

ARRECADACAO_GUICHE_SEM_FATURA = 'a'
"""a – Guichê de Caixa sem fatura/guia de arrecadação"""

ARRECADACAO_ELETRONICA_SEM_FATURA = 'b'
"""
b – Arrecadação Eletrônica sem fatura/guia de arrecadação (terminais de
auto - atendimento, ATM, home/office banking)
"""

ARRECADACAO_INTERNET_SEM_FATURA = 'c'
"""c – Internet sem fatura/guia de arrecadação"""

ARRECADACAO_CORRESPONDENTES_SEM_FATURA = 'd'
"""d – Correspondentes bancários sem fatura/guia de arrecadação"""

ARRECADACAO_TELEFONE_SEM_FATURA = 'e'
"""e – Telefone sem fatura/guia de arrecadação"""

ARRECADACAO_OUTROS_SEM_FATURA = 'f'
"""f – Outros meios sem fatura/guia de arrecadação"""

ARRECADACAO_LOTERICAS_SEM_FATURA = 'g'
"""g – Casas lotéricas sem fatura/guia de arrecadação"""

PAGAMENTO_DINHEIRO = '1'
"""1 – Dinheiro"""

PAGAMENTO_CHEQUE = '2'
"""2 – Cheque"""

PAGAMENTO_NAO_IDENTIFICADO = '3'
"""3 – Não identificado/outras formas"""

PAGAMENTO_OUTRAS_FORMAS = '3'
"""3 – Não identificado/outras formas"""



from datetime import date, datetime

from febraban_barcode import functions as fn

from .constants import (
    MODULO10_QUANTIDADE_MOEDA,
    MODULO10_VALOR_EFETIVO,
    MODULO11_QUANTIDADE_MOEDA,
    MODULO11_VALOR_EFETIVO,
    PRODUTO_ARRECADACAO,
    SEGMENTO_DEMAIS,
    SEGMENTO_ENERGIA_ELETRICA_GAS,
    SEGMENTO_EXCLUSIVO_BANCO,
    SEGMENTO_MULTAS_TRANSITO,
    SEGMENTO_ORGAOS_GOVERNAMENTAIS,
    SEGMENTO_PREFEITURA,
    SEGMENTO_SANEAMENTO,
    SEGMENTO_TELECOMUNICACOES,
)


def gerar_numeracao_codigo_de_barras(
    produto: int,
    segmento: int,
    codigo_moeda: int,
    valor: float,
    id_empresa_orgao: str,
    dados_campo_livre: str = '',
    vencimento: date | None = None,
    incluir_dac: bool = False,
) -> str:
    """
    Identificação da Empresa/Órgão:
        4  posições: Código Febraban ou código de compensação
        8  posições: (Empresa/Órgão) CNPJ ou primeiras oito posições do
        cadastro geral de contribuintes do Ministério da Fazenda
    """
    if id_empresa_orgao is None or len(id_empresa_orgao) not in (4, 8):
        raise Exception(
            'Identificação da Empresa/Órgão não informado ou inválido.'
        )

    numeracao = ''

    str_identificacao_produto = str(produto)
    str_identificacao_segmento = str(segmento)
    str_codigo_moeda = str(codigo_moeda)

    parte1 = (
        str_identificacao_produto
        + str_identificacao_segmento
        + str_codigo_moeda
    )

    str_valor = str(int(valor * 100)).zfill(11)

    """
    Se for utilizado o CNPJ para identificar a Empresa/Órgão, haverá uma
    redução no seu campo livre que passará a conter 21 posições.
    """
    str_campo_livre = ''

    if vencimento is not None:
        str_vencimento = vencimento.strftime('%Y%m%d')
    else:
        str_vencimento = ''
    zeros_totais = (21 if len(id_empresa_orgao) == 8 else 25) - len(
        str_vencimento
    )

    str_campo_livre = str_vencimento + dados_campo_livre.zfill(zeros_totais)

    parte2 = str_valor + id_empresa_orgao + str_campo_livre

    digito_verificador = ''

    if (
        codigo_moeda == MODULO10_VALOR_EFETIVO
        or codigo_moeda == MODULO10_QUANTIDADE_MOEDA
    ):
        digito_verificador = fn.calc_dac_modulo10(str(parte1) + str(parte2))
    elif (
        codigo_moeda == MODULO11_VALOR_EFETIVO
        or codigo_moeda == MODULO11_QUANTIDADE_MOEDA
    ):
        digito_verificador = fn.calc_dac_modulo11(str(parte1) + str(parte2))

    if digito_verificador == '':
        raise Exception('Dígito verificador inválido.')

    numeracao += parte1 + str(digito_verificador) + parte2

    if incluir_dac is False:
        return numeracao

    return calcular_dac(numeracao)


def calcular_dac(codigo_de_barras: str) -> str:
    codigo_de_barras_limpo = fn.limpar_codigo_de_barras(codigo_de_barras)
    if len(codigo_de_barras_limpo) != 44:
        raise Exception('Código de barra inválido.')

    codigo_moeda = int(codigo_de_barras_limpo[2])
    modulo = fn.dac_func(codigo_moeda)
    if modulo is None:
        raise Exception('Código moeda inválido.')

    str_linha_digitavel = ''

    partes = [
        codigo_de_barras_limpo[i : i + 11]
        for i in range(0, len(codigo_de_barras_limpo), 11)
    ]

    for parte in partes:
        str_linha_digitavel += parte + ' ' + str(modulo(parte)) + '   '

    return str_linha_digitavel.strip()


def decode_codigo_de_barras(
    codigo_de_barras: str, as_dict: bool = False
) -> None | dict:
    codigo_de_barras_limpo = fn.limpar_codigo_de_barras(codigo_de_barras)

    dac_blocos = []

    if len(codigo_de_barras_limpo) == 48:
        codigo_de_barras_digitos = [i for i in codigo_de_barras_limpo]
        dac_blocos.append(codigo_de_barras_digitos.pop(47))
        dac_blocos.append(codigo_de_barras_digitos.pop(35))
        dac_blocos.append(codigo_de_barras_digitos.pop(23))
        dac_blocos.append(codigo_de_barras_digitos.pop(11))
        dac_blocos.reverse()
        codigo_de_barras_44 = ''.join(codigo_de_barras_digitos)
    elif len(codigo_de_barras_limpo) == 44:
        codigo_de_barras_44 = codigo_de_barras_limpo
    else:
        raise Exception('Código de Barras inválido.')

    modulo = fn.dac_func(codigo_de_barras_44[2])

    codigo_de_barras_dict = {
        'codigo_de_barras_44': codigo_de_barras_44,
        'codigo_de_barras_desc': codigo_de_barras_44,
        'identificador_produto': '',
        'identificador_produto_desc': '',
        'identificador_segmento': '',
        'identificador_segmento_desc': '',
        'identificador_valor_ref': '',
        'identificador_valor_ref_desc': '',
        'digito_verificador': '',
        'digito_verificador_desc': '',
        'valor_efetivo_referencia': '',
        'valor_efetivo_referencia_desc': '',
        'identificador_empresa_orgao': '',
        'identificador_empresa_orgao_desc': '',
        'vencimento': None,
        'vencimento_desc': '',
        'campo_livre': '',
        'campo_livre_desc': '',
        'valido': True,
        'erro': [],
    }

    codigo_de_barras_str = 'Código de Barras: '
    partes = [
        codigo_de_barras_44[i : i + 11]
        for i in range(0, len(codigo_de_barras_44), 11)
    ]
    for i in range(len(partes)):
        dac = ''
        if len(dac_blocos) > 0:
            dac = '-' + dac_blocos[i]
        codigo_de_barras_str += partes[i] + dac + ' '
    codigo_de_barras_dict['codigo_de_barras_desc'] = codigo_de_barras_str

    if int(codigo_de_barras_44[0]) == PRODUTO_ARRECADACAO:
        codigo_de_barras_dict['identificador_produto'] = codigo_de_barras_44[0]
        codigo_de_barras_dict[
            'identificador_produto_desc'
        ] = 'Identificação do Produto: (8) Arrecadação.'
    else:
        raise Exception('Identificação do Produto inválido.')

    if int(codigo_de_barras_44[1]) == SEGMENTO_PREFEITURA:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (1) Prefeituras.'
    elif int(codigo_de_barras_44[1]) == SEGMENTO_SANEAMENTO:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (2) Saneamento.'
    elif int(codigo_de_barras_44[1]) == SEGMENTO_ENERGIA_ELETRICA_GAS:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (3) Energia Elétrica e Gás.'
    elif int(codigo_de_barras_44[1]) == SEGMENTO_TELECOMUNICACOES:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (4) Telecomunicações.'
    elif int(codigo_de_barras_44[1]) == SEGMENTO_ORGAOS_GOVERNAMENTAIS:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (5) Órgãos Governamentais.'
    elif int(codigo_de_barras_44[1]) == SEGMENTO_DEMAIS:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (6) Carnes e Assemelhados ou demais Empresas.'
    elif int(codigo_de_barras_44[1]) == SEGMENTO_MULTAS_TRANSITO:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (7) Multas de trânsito.'
    elif int(codigo_de_barras_44[1]) == SEGMENTO_EXCLUSIVO_BANCO:
        codigo_de_barras_dict['identificador_segmento'] = codigo_de_barras_44[
            1
        ]
        codigo_de_barras_dict[
            'identificador_segmento_desc'
        ] = 'Identificação do Segmento: (9) Uso exclusivo do banco.'
    else:
        raise Exception('Identificação do Segmento inválido.')

    if int(codigo_de_barras_44[2]) == MODULO10_VALOR_EFETIVO:
        codigo_de_barras_dict['identificador_valor_ref'] = codigo_de_barras_44[
            2
        ]
        codigo_de_barras_dict[
            'identificador_valor_ref_desc'
        ] = 'Identificador de Valor Efetivo ou Referência: (6) Valor a ser cobrado efetivamente em reais com dígito verificador calculado pelo módulo 10 na quarta posição do código de barras.'
    elif int(codigo_de_barras_44[2]) == MODULO10_QUANTIDADE_MOEDA:
        codigo_de_barras_dict['identificador_valor_ref'] = codigo_de_barras_44[
            2
        ]
        codigo_de_barras_dict[
            'identificador_valor_ref_desc'
        ] = 'Identificador de Valor Efetivo ou Referência: (7) Quantidade de moeda\n\tZeros – somente na impossibilidade de utilizar o valor;\n\tValor a ser reajustado por um índice\n\tcom dígito verificador calculado pelo módulo 10 na quarta posição do código de barras.'
    elif int(codigo_de_barras_44[2]) == MODULO11_VALOR_EFETIVO:
        codigo_de_barras_dict['identificador_valor_ref'] = codigo_de_barras_44[
            2
        ]
        codigo_de_barras_dict[
            'identificador_valor_ref_desc'
        ] = 'Identificador de Valor Efetivo ou Referência: (8) Valor a ser cobrado efetivamente em reais com dígito verificador calculado pelo módulo 11 na quarta posição do código de barras.'
    elif int(codigo_de_barras_44[2]) == MODULO11_QUANTIDADE_MOEDA:
        codigo_de_barras_dict['identificador_valor_ref'] = codigo_de_barras_44[
            2
        ]
        codigo_de_barras_dict[
            'identificador_valor_ref_desc'
        ] = 'Identificador de Valor Efetivo ou Referência: (9) Quantidade de moeda\n\tZeros – somente na impossibilidade de utilizar o valor; \n\tValor a ser reajustado por um índice\n\tcom dígito verificador calculado pelo módulo 11 na quarta posição do código de barras.'
    else:
        raise Exception(
            'Identificador de Valor Efetivo ou Referência inválido.'
        )

    digito_verificador_geral = codigo_de_barras_44[3]
    codigo_de_barras_dict['digito_verificador'] = digito_verificador_geral
    codigo_de_barras_dict['digito_verificador_desc'] = (
        'Dígito verificador geral (módulo 10 ou 11): '
        + digito_verificador_geral
    )

    dac = modulo(codigo_de_barras_44[0:3] + codigo_de_barras_44[4:44])
    if dac != int(digito_verificador_geral):
        codigo_de_barras_dict['erro'].append(
            'Erro: Dígito verificador geral inválido.'
        )
        codigo_de_barras_dict['valido'] = False

    valor_efetivo_ref = codigo_de_barras_44[4:15]

    if int(codigo_de_barras_44[2]) in (
        MODULO10_VALOR_EFETIVO,
        MODULO11_VALOR_EFETIVO,
    ):
        codigo_de_barras_dict['valor_efetivo_referencia'] = (
            float(valor_efetivo_ref) / 100
        )
        codigo_de_barras_dict[
            'valor_efetivo_referencia_desc'
        ] = 'Valor Efetivo: R$ ' + str(float(valor_efetivo_ref) / 100).replace(
            '.', ','
        )
    else:
        codigo_de_barras_dict['valor_efetivo_referencia'] = int(
            valor_efetivo_ref
        )
        codigo_de_barras_dict[
            'valor_efetivo_referencia_desc'
        ] = 'Valor Referência: ' + str(int(valor_efetivo_ref))

    identificador_empresa = codigo_de_barras_44[15:19]
    codigo_de_barras_dict[
        'identificador_empresa_orgao'
    ] = identificador_empresa

    if int(codigo_de_barras_44[1]) == SEGMENTO_EXCLUSIVO_BANCO:
        codigo_de_barras_dict['identificador_empresa_orgao_desc'] = (
            'Identificação da Empresa/Órgão: '
            + identificador_empresa
            + ' código de compensação.'
        )
    else:
        codigo_de_barras_dict['identificador_empresa_orgao_desc'] = (
            'Identificação da Empresa/Órgão: '
            + identificador_empresa
            + ' código de Febraban.'
        )

    campo_livre = codigo_de_barras_44[19:44]

    try:
        vencimento = datetime.strptime(campo_livre[0:8], '%Y%m%d')
        codigo_de_barras_dict['vencimento'] = vencimento.strftime('%Y-%m-%d')
        codigo_de_barras_dict[
            'vencimento_desc'
        ] = 'Data de Vencimento: ' + vencimento.strftime('%d/%m/%Y')
    except Exception:
        codigo_de_barras_dict[
            'vencimento_desc'
        ] = 'Data de Vencimento: Não informado.'

    codigo_de_barras_dict['campo_livre'] = campo_livre
    codigo_de_barras_dict['campo_livre_desc'] = (
        'Campo Livre (25 posições): ' + campo_livre
    )

    partes = [
        codigo_de_barras_44[i : i + 11]
        for i in range(0, len(codigo_de_barras_44), 11)
    ]

    if len(codigo_de_barras_limpo) == 48:
        for i in range(len(partes)):
            if modulo(partes[i]) != int(dac_blocos[i]):
                codigo_de_barras_dict['erro'].append(
                    'Erro: DAC (Dígito de Auto-Conferência) do bloco '
                    + str(i + 1)
                    + ' inválido.'
                )
                codigo_de_barras_dict['valido'] = False

    if as_dict:
        return {
            key: value
            for (key, value) in codigo_de_barras_dict.items()
            if '_desc' not in key
        }

    for key, desc in codigo_de_barras_dict.items():
        if '_desc' in key:
            print(desc)

    if not codigo_de_barras_dict['valido']:
        print('STATUS: Inválido.')
        for erro in codigo_de_barras_dict['erro']:
            print(' -> ' + erro)
    else:
        print('STATUS: Válido.')



from datetime import datetime

from febraban_barcode import constants as c


def registro_A(
    codigo_remessa='2',
    codigo_convenio=' ',
    nome_empresa=' ',
    codigo_banco='0',
    nome_banco=' ',
    data_geracao='',
    nsa='0',
    versao_layout='05',
    servico='CODIGO DE BARRAS',
    reservado='',
):
    """
    REGISTRO “A” – HEADER

    Args:
        codigo_remessa (str, optional): A.02 - Código de Remessa
            2 - RETORNO - Enviado pelo Banco para a Empresa/Órgão. Defaults to '2'.
        codigo_convenio (str, optional): A.03 - Código do Convênio
            Definido pelo banco. Defaults to ' '.
        nome_empresa (str, optional): A.04 - Nome da Empresa/Órgão. Defaults to ' '.
        codigo_banco (str, optional): A.05 - Código do Banco
            Código do Banco na câmara de compensação. Defaults to '0'.
        nome_banco (str, optional): A.06 - Nome do Banco. Defaults to ' '.
        data_geracao (str, optional): A.07 - Data da geração do arquivo (AAAAMMDD). Defaults to '0'.
        nsa (str, optional): A.08 - Número seqüencial do arquivo ( NSA )
            Este número deverá evoluir de 1 em 1 para cada arquivo gerado. Defaults to '0'.
        versao_layout (str, optional): A.09 - Versão do lay - out
            atual = versão 04
            nova = versão 05 - disponível a partir de 01.08.2016. Defaults to '05'.
        servico (str, optional): A.10 - Identificação do serviço
            Deverá conter “CÓDIGO DE BARRAS”. Defaults to 'CODIGO DE BARRAS'.
        reservado (str, optional): A.11 - Reservado para o futuro (filler). Defaults to ' '.

    Raises:
        Exception: Campo inválido.

    Returns:
        str: Retorna o header com 150bytes.
    """
    registro_a = 'A'

    if len(codigo_remessa) != 1:
        raise Exception('Campo A.02 Inválido. O campo deve ter o tamanho 1.')
    if not codigo_remessa.isdigit():
        raise Exception('Campo A.02 Inválido. O campo deve ser numérico.')
    registro_a += codigo_remessa

    if len(codigo_convenio) > 20:
        raise Exception(
            'Campo A.03 Inválido. O campo deve ter o tamanho máximo de 20.'
        )
    registro_a += codigo_convenio.ljust(20)

    if len(nome_empresa) > 20:
        raise Exception(
            'Campo A.04 Inválido. O campo deve ter o tamanho máximo de 20.'
        )
    registro_a += nome_empresa.ljust(20)

    if len(codigo_banco) > 3:
        raise Exception(
            'Campo A.05 Inválido. O campo deve ter o tamanho máximo de 3.'
        )
    if not codigo_banco.isdigit():
        raise Exception('Campo A.05 Inválido. O campo deve ser numérico.')
    registro_a += codigo_banco.zfill(3)

    if len(nome_banco) > 20:
        raise Exception(
            'Campo A.06 Inválido. O campo deve ter o tamanho máximo de 20.'
        )
    registro_a += nome_banco.upper().ljust(20)

    try:
        datetime.strptime(data_geracao, '%Y%m%d')
    except:
        raise Exception(
            'Campo A.07 Inválido. O campo deve ser uma data no formato AAAAMMDD.'
        )
    registro_a += data_geracao

    if len(nsa) > 6:
        raise Exception(
            'Campo A.08 Inválido. O campo deve ter o tamanho máximo de 6.'
        )
    if not nsa.isdigit():
        raise Exception('Campo A.08 Inválido. O campo deve ser numérico.')
    registro_a += nsa.zfill(6)

    if len(versao_layout) > 2:
        raise Exception(
            'Campo A.09 Inválido. O campo deve ter o tamanho máximo de 2.'
        )
    if not versao_layout.isdigit():
        raise Exception('Campo A.09 Inválido. O campo deve ser numérico.')
    registro_a += versao_layout.zfill(2)

    if len(servico) > 17:
        raise Exception(
            'Campo A.10 Inválido. O campo deve ter o tamanho máximo de 17.'
        )
    registro_a += servico.upper().ljust(17)

    if len(reservado) > 52:
        raise Exception(
            'Campo A.11 Inválido. O campo deve ter o tamanho máximo de 52.'
        )
    registro_a += reservado.upper().ljust(52)
    return registro_a


def registro_G(
    codigo_banco='0',
    agencia='0',
    conta='0',
    conta_digito='0',
    data_pagamento='',
    data_credito='',
    codigo_de_barras_44=' ',
    valor_recebido=0.0,
    valor_tarifa=0.0,
    nsr=None,
    codigo_agencia=' ',
    forma_arrecadacao=' ',
    autenticacao=' ',
    forma_pagamento='0',
    reservado='',
):
    """
    REGISTRO “G” – RETORNO DAS ARRECADAÇÕES IDENTIFICADAS COM CÓDIGO DE BARRAS

    Args:
        codigo_banco (str, optional): G.02 - Identificação do banco da empresa/órgão creditada. Defaults to '0'.
        agencia (str, optional): G.02 - Identificação da agência da empresa/órgão creditada. Defaults to '0'.
        conta (str, optional): G.02 - Identificação da conta da empresa/órgão creditada. Defaults to '0'.
        conta_digito (str, optional): G.02 - Identificação do dígito da conta da empresa/órgão creditada. Defaults to '0'.
        data_pagamento (str, optional): G.03 - Data de pagamento (AAAA/MM/DD). Defaults to ''.
        data_credito (str, optional): G.04 - Data de crédito (AAAA/MM/DD). Defaults to ''.
        codigo_de_barras_44 (str, optional): G.05 - Código de Barras (44 posições). Defaults to ' '.
        valor_recebido (float, optional): G.06 - Valor recebido. Defaults to 0.0.
        valor_tarifa (float, optional): G.07 - Valor da tarifa referente a cada comprovante arrecadado (será informado desde que acordado entre as partes). Defaults to 0.0.
        nsr (_type_, optional): G.08 - NSR - Número Seqüencial de Registro. Defaults to None.
        codigo_agencia (str, optional): G.09 - Código da agência arrecadadora. Defaults to ' '.
        forma_arrecadacao (str, optional): G.10 – Forma de arrecadação/captura (canais de recebimento)
            1 – Guichê de Caixa com fatura/guia de arrecadação
            2 – Arrecadação Eletrônica com fatura/guia de arrecadação (terminais de auto - atendimento, ATM, home/office banking)
            3 – Internet com fatura/guia de arrecadação
            4 – Outros meios com fatura/guia de arrecadação
            5 – Correspondentes bancários com fatura/guia de arrecadação
            6 – Telefone com fatura/guia de arrecadação
            7 – Casas lotéricas com fatura/guia de arrecadação
            a – Guichê de Caixa sem fatura/guia de arrecadação
            b – Arrecadação Eletrônica sem fatura/guia de arrecadação (terminais de auto - atendimento, ATM, home/office banking)
            c – Internet sem fatura/guia de arrecadação
            d – Correspondentes bancários sem fatura/guia de arrecadação
            e – Telefone sem fatura/guia de arrecadação
            f – Outros meios sem fatura/guia de arrecadação
            g – Casas lotéricas sem fatura/guia de arrecadação. Defaults to ' '.
        autenticacao (str, optional): G.11 – Número de autenticação caixa ou código de transação (será informado desde que acordado entre as partes). Defaults to ' '.
        forma_pagamento (str, optional): G.12 – Forma de Pagamento
            1 – Dinheiro
            2 – Cheque
            3 – Não identificado/outras formas. Defaults to '0'.
        reservado (str, optional): G.13 – Reservado para o futuro. Defaults to ' '.

    Raises:
        Exception: Campo inválido.

    Returns:
        str: Retorna um registro com 150bytes.
    """
    registro_g = 'G'

    if len(codigo_banco) > 3:
        raise Exception(
            'Campo G.02 Inválido. O campo deve ter o tamanho máximo de 3.'
        )
    if not codigo_banco.isdigit():
        raise Exception(
            'Campo G.02 Inválido. O campo codigo_banco deve ser numérico.'
        )
    registro_g += codigo_banco.zfill(3)

    if not agencia.isdigit():
        raise Exception(
            'Campo G.02 Inválido. O campo agencia deve ser numérico.'
        )

    if not conta.isdigit():
        raise Exception(
            'Campo G.02 Inválido. O campo conta deve ser numérico.'
        )

    if not conta_digito.isdigit():
        raise Exception(
            'Campo G.02 Inválido. O campo conta_digito deve ser numérico.'
        )

    agencia_conta = agencia + conta + conta_digito
    if len(agencia_conta) > 17:
        raise Exception(
            'Campo G.02 Inválido. Os campos agencia, conta e conta_digito juntos devem ter tamanho máximo de 17.'
        )
    registro_g += agencia_conta.zfill(17)

    try:
        datetime.strptime(data_pagamento, '%Y%m%d')
    except:
        raise Exception(
            'Campo G.03 Inválido. O campo deve ser uma data no formato AAAAMMDD.'
        )
    registro_g += data_pagamento

    try:
        datetime.strptime(data_credito, '%Y%m%d')
    except:
        raise Exception(
            'Campo G.04 Inválido. O campo deve ser uma data no formato AAAAMMDD.'
        )
    registro_g += data_credito

    if len(codigo_de_barras_44) != 44:
        raise Exception(
            'Campo G.05 Inválido. O campo deve ser um código de barras de 44 posições.'
        )
    registro_g += codigo_de_barras_44

    if not isinstance(valor_recebido, float):
        raise Exception('Campo G.06 Inválido. O campo deve ser um float.')
    registro_g += str(int(valor_recebido * 100)).zfill(12)

    if not isinstance(valor_tarifa, float):
        raise Exception('Campo G.07 Inválido. O campo deve ser um float.')
    registro_g += str(int(valor_tarifa * 100)).zfill(7)

    if not isinstance(nsr, int):
        raise Exception('Campo G.08 Inválido. O campo deve ser um int.')
    registro_g += str(nsr).zfill(8)

    if len(codigo_agencia) > 8:
        raise Exception(
            'Campo G.09 Inválido. O campo deve ter o tamanho máximo de 8.'
        )
    registro_g += codigo_agencia.ljust(8)

    if len(forma_arrecadacao) > 1:
        raise Exception('Campo G.10 Inválido. O campo deve ter o tamanho 1.')

    if forma_arrecadacao not in [
        c.ARRECADACAO_GUICHE_COM_FATURA,
        c.ARRECADACAO_ELETRONICA_COM_FATURA,
        c.ARRECADACAO_INTERNET_COM_FATURA,
        c.ARRECADACAO_OUTROS_COM_FATURA,
        c.ARRECADACAO_CORRESPONDENTES_COM_FATURA,
        c.ARRECADACAO_TELEFONE_COM_FATURA,
        c.ARRECADACAO_LOTERICAS_COM_FATURA,
        c.ARRECADACAO_GUICHE_SEM_FATURA,
        c.ARRECADACAO_ELETRONICA_SEM_FATURA,
        c.ARRECADACAO_INTERNET_SEM_FATURA,
        c.ARRECADACAO_CORRESPONDENTES_SEM_FATURA,
        c.ARRECADACAO_TELEFONE_SEM_FATURA,
        c.ARRECADACAO_OUTROS_SEM_FATURA,
        c.ARRECADACAO_LOTERICAS_SEM_FATURA,
    ]:
        raise Exception('Campo G.10 Inválido.')

    registro_g += forma_arrecadacao.ljust(1)

    if len(autenticacao) > 23:
        raise Exception(
            'Campo G.11 Inválido. O campo deve ter o tamanho máximo de 23.'
        )
    registro_g += autenticacao.ljust(23)

    if len(forma_pagamento) > 1:
        raise Exception(
            'Campo G.12 Inválido. O campo deve ter o tamanho máximo de 1.'
        )
    if not forma_pagamento.isdigit():
        raise Exception('Campo G.12 Inválido. O campo deve ser numérico.')
    registro_g += forma_pagamento.zfill(1)

    if len(reservado) > 9:
        raise Exception(
            'Campo G.13 Inválido. O campo deve ter o tamanho máximo de 9.'
        )
    registro_g += reservado.upper().ljust(9)
    return registro_g


def registro_Z(total_registros=0, valor_total_registros=0.0, reservado=''):
    """
    REGISTRO “Z” - TRAILLER

    Args:
        total_registros (int, optional): Z.02 - Total de registros no arquivo
            Total de registros no arquivo, inclusive com header e trailler. Defaults to 0.
        valor_total_registros (float, optional): Z.03 - Valor total dos registros do arquivo. Defaults to 0.0.
        reservado (str, optional): Z.04 - Reservado para o futuro (filler). Defaults to ''.

    Raises:
        Exception: Campo inválido.

    Returns:
        str: Retorna um registro com 150bytes.
    """
    registro_z = 'Z'

    if not isinstance(total_registros, int):
        raise Exception('Campo Z.02 Inválido. O campo deve ser um int.')
    registro_z += str(total_registros).zfill(6)

    if not isinstance(valor_total_registros, float):
        raise Exception('Campo Z.03 Inválido. O campo deve ser um float.')
    registro_z += str(int(valor_total_registros * 100)).zfill(17)

    if len(reservado) > 126:
        raise Exception(
            'Campo Z.04 Inválido. O campo deve ter o tamanho máximo de 126.'
        )
    registro_z += reservado.upper().ljust(126)
    return registro_z
